# coding=utf-8
#  Copyright (c) 2019 ByteDance Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import re


def is_kernal_stack_line(stack_line=''):
    return stack_line.strip().startswith('kernal')


def is_java_layer_stack_line(stack_line=''):
    return stack_line.strip().startswith('at ')


def is_lock_obj_stack_line(stack_line=''):
    return stack_line.strip().startswith('- locked')


def is_aosp_code_stack_line(stack_line=''):
    strip_line = stack_line.strip()
    return strip_line.startswith('at android.') or \
           strip_line.startswith('at com.android.internal') or \
           strip_line.startswith('at com.android.app') or \
           strip_line.startswith('at java.lang')


def is_binder_transact(stack_line=''):
    return stack_line.strip().startswith('at android.os.BinderProxy.')


def is_binder_transact_native(stack_line=''):
    return stack_line.__contains__('/system/lib/libbinder.so')


def is_binder_invoke_stack(stack_tree):
    for stack_line in stack_tree:
        if is_binder_transact_native(stack_line) or is_binder_transact(stack_line):
            return True
    return False


# at android.app.IActivityManager$Stub$Proxy.activityPaused(IActivityManager.java:4596)
BINDER_INVOKE_METHOD = re.compile(r'([^.]*)\(')
BINDER_INTERFACE_NAME = re.compile(r'\.([^.$]*)\$')


def get_binder_invoke_method(stack_tree):
    for stack_line in stack_tree:
        if is_java_layer_stack_line(stack_line) and not is_binder_transact(stack_line):
            finders = BINDER_INVOKE_METHOD.findall(stack_line)
            method_name = finders[0]
            finders = BINDER_INTERFACE_NAME.findall(stack_line)
            interface_name = finders[0]
            return interface_name, method_name


def parse_native_stack(basic_cause, stack_tree):
    # 将前3行native堆栈输出
    first_three_stack = get_some_stack(stack_tree)
    basic_cause += ("最后调用在Native代码:\n%s\n" % first_three_stack)

    # 将调用native的关键java层打出
    java_layer_stack_line = get_some_stack(stack_tree, is_java_layer_stack_line)
    if java_layer_stack_line:
        basic_cause += "Java层触发关键:\n%s\n" % java_layer_stack_line

    # 将上层应用的可以第一个堆栈输出
    not_aosp_stack_line = get_some_stack(stack_tree, is_java_no_aosp_code_line, 2)
    if not_aosp_stack_line:
        basic_cause += "上层调用关键:\n%s\n" % not_aosp_stack_line
    else:
        basic_cause += "关键堆栈都是AOSP的调用栈\n"

    return basic_cause


def is_java_no_aosp_code_line(stack_line):
    return is_java_layer_stack_line(stack_line) and not is_aosp_code_stack_line(stack_line)


class BinderStackHelper:
    def __init__(self):
        pass

    interface_name = None
    method_name = None
    is_binder_invoke = False
    guess_ref_class_name = None
    server_invoke_stack_tree = None
    server_invoke_process_all_data = None

    def process(self, stack_tree, is_top_main_process):
        if is_top_main_process and is_binder_invoke_stack(stack_tree):
            self.interface_name, self.method_name = get_binder_invoke_method(stack_tree)
            self.is_binder_invoke = True

            if self.interface_name.strip().startswith("I"):
                self.guess_ref_class_name = self.interface_name[1:]
            return True
        else:
            return self.is_binder_invoke and self.is_ref_stack_tree(stack_tree)

    def process_process_all_data(self, process_data):
        if self.server_invoke_stack_tree is not None and self.server_invoke_process_all_data is None:
            if process_data.is_stack_belong(self.server_invoke_stack_tree):
                self.server_invoke_process_all_data = process_data

    def is_ref_stack_tree(self, stack_tree):
        if not self.is_binder_invoke:
            return False

        for stack_line in stack_tree:
            if self.interface_name in stack_line:
                self.server_invoke_stack_tree = stack_tree
                return True
            if self.method_name in stack_line:
                return True
            if self.guess_ref_class_name is not None and self.guess_ref_class_name in stack_line:
                return True

        return False

    def guess_service_invoke_detail(self, basic_cause):
        if self.server_invoke_stack_tree is None:
            basic_cause += "但是未找到有效对端调用堆栈，下面相关堆栈中将打出所有相关可以参考\n"
        else:
            head_info = self.server_invoke_process_all_data.head_info
            cmd_line = self.server_invoke_process_all_data.cmd_line
            thread_info_stacks = self.server_invoke_process_all_data.thread_info_stacks
            basic_cause += "Binder对端[%s, %s]线程[%s, %s]调用:\n" % \
                           (cmd_line.to_string(), head_info.pid, thread_info_stacks.thread_name, thread_info_stacks.tid)
            # 给三行堆栈
            stack_tree = thread_info_stacks.stack_tree
            some_stack = get_some_stack(stack_tree, is_java_layer_stack_line)
            basic_cause += some_stack

        return basic_cause


def get_some_stack(stack_tree, condition_func=None, line_count=3):
    stacks = None
    stack_line_count = 0
    for stack_line in stack_tree:
        if stacks is None and callable(condition_func):
            pre_condition = condition_func(stack_line)
        else:
            pre_condition = True

        if pre_condition and stacks is None:
            stacks = ''

        if pre_condition:
            stacks += stack_line + "\n"

        if stacks is not None:
            stack_line_count += 1

        if stack_line_count >= line_count:
            break

    return stacks
